<?php
/*
  STORM is under the MIT License (MIT)

  Copyright (c) 2010-2022 Agence Française Informatique http://www.afi-sa.fr

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
*/


class Storm_Test_LoaderQueryVolatileClauseWhereTest
  extends Storm_Test_ModelTestCase {

  protected array $_storm_scopes = ['LoaderQueryVolatileClauseWhereTest'];

  public function setUp(): void
  {
    parent::setUp();

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 100,
                    'login' => 'user_admin',
                    'level' => 'invite admin redacteur',
                    'foo' => 'premier deuxieme troisieme',
                    'prefs' => null,
                    'sys' => '',
                    'date' => '2022-01-01']);

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 101,
                    'login' => 'user_administrateur',
                    'level' => 'administrateur',
                    'foo' => 'deuxieme redacteur',
                    'prefs' => '{"best":true}',
                    'sys' => null,
                    'date' => '2020-01-01']);

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 102,
                    'login' => 'user_invite',
                    'level' => 'invite',
                    'foo' => 'premier deuxieme',
                    'url' => 'https://my-server.org/page_1.html',
                    'website' => 'https://my-server.org/pomme+d\'api.php page2.php',
                    'sys' => 'content',
                    'date' => '2021-01-01']);
  }


  /** @test */
  public function withLimit_0_ShouldReturnNothing() {
    $this->assertEmpty(Storm_Test_LoaderQueryUser::query()
                       ->limit(0)
                       ->fetchAll());
  }


  /** @test */
  public function withTwoFollowingFetchAllQueryShouldAnswersCorrectLoader() {
    $query = Storm_Test_LoaderQueryUser::query();

    $results = $query->fetchAll();
    $this->assertTrue($results[0] instanceof Storm_Test_LoaderQueryUser);

    $results = $query->fetchAll();
    $this->assertTrue($results[0] instanceof Storm_Test_LoaderQueryUser);
  }


  /** @test */
  public function withOnlyOneClauseInWrapCriteriaAndShouldNotStartWhereWithAnd() {
    $query = Storm_Test_LoaderQueryUser::query();

    $query->wrapWithAnd((new Storm_Query_Criteria)
                        ->eq('id', 100))
          ->fetchAll();

    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`id` = 100)');
  }


  /** @test */
  public function queryWrapShouldChangeOperator() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->eq('id', 100)
      ->gt('id', 101);

    $query->wrapWithOr((new Storm_Query_Criteria)
                       ->match((new Storm_Query_MatchRating('foo'))
                               ->against_or('premier')))
          ->fetchAll();

    $query->wrapWithAnd((new Storm_Query_Criteria)
                        ->start('sys', 'cont'))
          ->fetchAll();

    $query->append((new Storm_Query_Criteria)
                   ->like('foo', '%deux%')
                   ->not_eq('level', 'invite'))
          ->fetchAll();

    $this->assertSqlEquals(['SELECT `users`.* FROM `users` WHERE ((`users`.`id` = 100 AND `users`.`id` > 101) OR MATCH(`users`.`foo`) AGAINST(\'premier\'))',
                            'SELECT `users`.* FROM `users` WHERE (((`users`.`id` = 100 AND `users`.`id` > 101) OR MATCH(`users`.`foo`) AGAINST(\'premier\')) AND `users`.`sys` LIKE \'cont%\')',
                            'SELECT `users`.* FROM `users` WHERE (((`users`.`id` = 100 AND `users`.`id` > 101) OR MATCH(`users`.`foo`) AGAINST(\'premier\')) AND `users`.`sys` LIKE \'cont%\' AND (`users`.`foo` LIKE \'%deux%\' AND `users`.`level` != \'invite\'))']);
  }


  /** @test */
  public function queryCheckHasOrders() {
    $this->assertTrue(Storm_Test_LoaderQueryUser::query()
                      ->order('login')
                      ->hasOrders());

    $this->assertFalse(Storm_Test_LoaderQueryUser::query()
                       ->hasOrders());
  }


  /** @test */
  public function queryCheckHasClauses() {
    $this->assertTrue(Storm_Test_LoaderQueryUser::query()
                      ->eq('id', 100)
                      ->hasClauses());
    $this->assertTrue(Storm_Test_LoaderQueryUser::join('tab1')
                      ->inner(Storm_Test_LoaderQueryUser::inner('tab2')
                              ->on_eq('id', 'id')
                              ->gt('id', 100))
                      ->hasClauses());

    $this->assertFalse(Storm_Test_LoaderQueryUser::query()
                       ->hasClauses());
    $this->assertFalse(Storm_Test_LoaderQueryUser::query()
                       ->limit(5)
                       ->hasClauses());
    $this->assertFalse(Storm_Test_LoaderQueryUser::query()
                       ->limit_page(5, 50)
                       ->hasClauses());
    $this->assertFalse(Storm_Test_LoaderQueryUser::query()
                       ->group('id')
                       ->hasClauses());
    $this->assertFalse(Storm_Test_LoaderQueryUser::join('tab1')
                       ->inner(Storm_Test_LoaderQueryUser::inner('tab2')
                               ->on_eq('id', 'id'))
                       ->hasClauses());
  }


  /** @test */
  public function whereId_GT_100_ShouldAnswersUserAdministrateurAndUserInvite() {
    $this->assertEquals(['user_administrateur', 'user_invite'],
                        (new Storm_Model_Collection(Storm_Test_LoaderQueryUser::query()
                                                    ->gt('id', 100)
                                                    ->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`id` > 100)');
  }


  /** @test */
  public function whereId_GT_EQ_101_ShouldAnswersUserAdministrateurAndUserInvite() {
    $this->assertEquals(['user_administrateur', 'user_invite'],
                        (new Storm_Model_Collection(Storm_Test_LoaderQueryUser::query()
                                                    ->gt_eq('id', 101)
                                                    ->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`id` >= 101)');
  }


  /** @test */
  public function whereId_LT_102_ShouldAnswersUserAdminAndUserAdministrateur() {
    $this->assertEquals(['user_admin', 'user_administrateur'],
                        (new Storm_Model_Collection(Storm_Test_LoaderQueryUser::query()
                                                    ->lt('id', 102)
                                                    ->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSqlCount('SELECT `users`.* FROM `users` WHERE (`users`.`id` < 102)', 1);
  }


  /** @test */
  public function whereId_LT_EQ_101_ShouldAnswersUserAdminAndUserAdministrateur() {
    $this->assertEquals(['user_admin', 'user_administrateur'],
                        (new Storm_Model_Collection(Storm_Test_LoaderQueryUser::query()
                                                    ->lt_eq('id', 101)
                                                    ->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`id` <= 101)');
    $this->assertNotSql('SELECT `users`.* FROM `users` WHERE (`users`.`id` > 101)');
  }


  /** @test */
  public function whereLogin_NOT_EQ_UserAdministrateurShouldReturnUserAdminAndUserInvite() {
    $this->assertEquals(['user_admin', 'user_invite'],
                        (new Storm_Model_Collection(Storm_Test_LoaderQueryUser::query()
                                                    ->not_eq('login', 'user_administrateur')
                                                    ->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`login` != \'user_administrateur\')');
  }


  /** @test */
  public function whereId_IN_100And102_AndFoo_NOT_LIKE_TroisieShouldReturnId_102() {
    $this->assertEquals([102],
                        (new Storm_Model_Collection(Storm_Test_LoaderQueryUser::query()
                                                    ->in('id', [100, 102])
                                                    ->not_like('foo', '%troisie%')
                                                    ->fetchAll()))
                        ->collect('id')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`id` IN (100, 102) AND `users`.`foo` NOT LIKE \'%troisie%\')');
  }


  /** @test */
  public function whereId_IN_100And101_AndFoo_START_DeuxShouldReturnId_101() {
    $this->assertEquals([101],
                        (new Storm_Model_Collection(Storm_Test_LoaderQueryUser::query()
                                                    ->in('id', [100, 101])
                                                    ->start('foo', 'deux')
                                                    ->fetchAll()))
                        ->collect('id')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`id` IN (100, 101) AND `users`.`foo` LIKE \'deux%\')');
  }


  /** @test */
  public function whereNotFoo_START_DeuxShouldReturnIds_100_102() {
    $this->assertEquals([100, 102],
                        (new Storm_Model_Collection(Storm_Test_LoaderQueryUser::query()
                                                    ->not_start('foo', 'deux')
                                                    ->fetchAll()))
                        ->collect('id')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`foo` NOT LIKE \'deux%\')');
  }


  /** @test */
  public function whereFoo_END_ActeurShouldReturnId_101() {
    $this->assertEquals([101],
                        (new Storm_Model_Collection(Storm_Test_LoaderQueryUser::query()
                                                    ->end('foo', 'acteur')
                                                    ->fetchAll()))
                        ->collect('id')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`foo` LIKE \'%acteur\')');
  }


  /** @test */
  public function whereNotFoo_END_ActeurShouldReturnId_100_102() {
    $this->assertEquals([100, 102],
                        (new Storm_Model_Collection(Storm_Test_LoaderQueryUser::query()
                                                    ->not_end('foo', 'acteur')
                                                    ->fetchAll()))
                        ->collect('id')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`foo` NOT LIKE \'%acteur\')');
  }


  /** @test */
  public function matchWithAgainstOn_0_ShouldReturnQueryWithAllAgainst() {
    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 103,
                    'level' => '1 2 3']);
    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 104,
                    'level' => '9 0 8']);

    $result = Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level'))
              ->against_or('1 0'))
      ->fetchAll();

    $this->assertEquals([103, 104],
                        (new Storm_Model_Collection($result))
                        ->collect('id')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`) AGAINST(\'1 0\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function matchWithAndExactAndAgainstAndExactShouldReturnExpected() {
    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 103,
                    'level' => '1 2 3']);
    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 104,
                    'level' => '1 3 2']);

    $result = Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level'))
              ->exact_and('1 2')
              ->against_and('3')
              ->exact_and('2 3'))
      ->fetchAll();

    $this->assertEquals([103],
                        (new Storm_Model_Collection($result))
                        ->collect('id')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`) AGAINST(\'+"1 2" +3 +"2 3"\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withMatchLevelAgainstAdminShouldAnswersOnlyUser_Admin() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level'))
              ->against_or('admin'));

    $this->assertEquals(['user_admin'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`) AGAINST(\'admin\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withNotMatchLevelAgainstAdminShouldAnswersUsers_Administrateur_Invite() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->not_match((new Storm_Query_MatchBoolean('level'))
                  ->against_or(['admin']));

    $this->assertEquals(['user_administrateur', 'user_invite'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (NOT MATCH(`users`.`level`) AGAINST(\'admin\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withMatchLevelAgainstAdminDeuxiemeShouldAnswersUser_AdminAndUser_AdministrateurAndUser_Invite() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level,foo'))
              ->against_or(['admin', 'deuxieme']));

    $this->assertEquals(['user_admin', 'user_administrateur', 'user_invite'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`, `users`.`foo`) AGAINST(\'admin deuxieme\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withMatchLevelAgainstNotAdminDeuxiemeShouldAnswersId_101() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level,foo'))
              ->against_not(['admin', 'premier']));

    $this->assertEquals([101],
                        array_map(fn($row) => $row->getId(),
                                  $query->fetchAll()));
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`, `users`.`foo`) AGAINST(\'-(admin premier)\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withMatchLevelAgainstAdmin_StarShouldAnswersId_100_101() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level'))
              ->against_or('admin*'));

    $this->assertEquals([100, 101],
                        array_map(fn($row) => $row->getId(),
                                  $query->fetchAll()));
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`) AGAINST(\'admin*\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withMatchLevelAgainstStar_ActeurShouldAnswersId_101() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('foo'))
              ->against_or('*acteur'));

    $this->assertEquals([101],
                        array_map(fn($row) => $row->getId(),
                                  $query->fetchAll()));
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`foo`) AGAINST(\'*acteur\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withMatchLevelAgainstStrictAdminRedacteur_DeuxiemeShouldAnswersUser_AdminAndUser_Administrateur() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level,foo'))
              ->against_and(['admin', 'redacteur'])
              ->against_and(['deuxieme']));

    $this->assertEquals(['user_admin', 'user_administrateur'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`, `users`.`foo`) AGAINST(\'+(admin redacteur) +deuxieme\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withMatchLevelFooExactDeuxiemeRedacteurShouldAnswersOnlyUser_Administrateur() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level,foo'))
              ->exact_or('deuxieme redacteur'));

    $this->assertEquals(['user_administrateur'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`, `users`.`foo`) AGAINST(\'"deuxieme redacteur"\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withMatchLevelFooDeuxiemeAndAgainstRedacteurAnswers_AllUsers() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level,foo'))
              ->against_or('deuxieme')
              ->against_or('redacteur'));

    $this->assertEquals(['user_admin', 'user_administrateur', 'user_invite'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`, `users`.`foo`) AGAINST(\'deuxieme redacteur\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withMatchLevelFooDeuxiemeAndAgainstStrictRedacteurAnswers_User_Admin_User_Administrateur() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level,foo'))
              ->against_or('deuxieme')
              ->against_and('redacteur'));

    $this->assertEquals(['user_admin', 'user_administrateur'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`, `users`.`foo`) AGAINST(\'deuxieme +redacteur\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withMatchWithMultipleTermsShouldReturn_Admin_Invite()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level,foo'))
              ->against_and([Storm_Query_Terms::and(['admin']),
                             Storm_Query_Terms::and(['administrateurs', 'administration'])])
              ->against_or(['redacteurs', 'premier']));

    $logins = (new Storm_Model_Collection($query->fetchAll()))
      ->collect('login')
      ->getArrayCopy();
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`, `users`.`foo`) AGAINST(\'(+admin +(administrateurs administration)) redacteurs premier\' IN BOOLEAN MODE))');
    $this->assertEquals(['user_admin', 'user_invite'], $logins);
  }


  /** @test */
  public function withMatchWithMultipleTermsComplexShouldReturn_Admin_Invite()
  {
    $query = Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level,foo'))
              ->against_and([Storm_Query_Terms::or([Storm_Query_Terms::and(['admin', 'administration']),
                                                    'invite'])]));

    $logins = (new Storm_Model_Collection($query->fetchAll()))
      ->collect('login')
      ->getArrayCopy();
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`, `users`.`foo`) AGAINST(\'+((+(admin administration)) invite)\' IN BOOLEAN MODE))');
    $this->assertEquals(['user_admin', 'user_invite'], $logins);
  }


  /** @test */
  public function withOrMatchWithMultipleTermsAndComplexShouldReturnSql()
  {
    Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level,foo'))
              ->against_or([Storm_Query_Terms::and([Storm_Query_Terms::and(['admin', 'administration']),
                                                    Storm_Query_Terms::and(['invite', 'premier'])]),
                            Storm_Query_Terms::and([Storm_Query_Terms::and(['admin2', 'administration2']),
                                                    Storm_Query_Terms::and(['invite2', 'premier2'])])]))
      ->fetchAll();

    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`, `users`.`foo`) AGAINST(\'(+(admin administration) +(invite premier)) (+(admin2 administration2) +(invite2 premier2))\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withMatchWithAndTermsInsideOrShouldReturnSql()
  {
    Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level'))
              ->against_or([Storm_Query_Terms::and(['administrateurs', 'administration']),
                            Storm_Query_Terms::and(['invite', 'premier'])]))
      ->fetchAll();

    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`) AGAINST(\'(+(administrateurs administration)) (+(invite premier))\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withSameDoubleOperatorShouldReturnSqlQueryWithOnlyOneOperator() {
    Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level'))
              ->against_and(Storm_Query_Terms::and(['admin'])))
      ->fetchAll();

    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`) AGAINST(\'+admin\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withDifferentDoubleOperatorDistinctShouldReturnSqlQueryWithTwoOperator() {
    $result = Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('level'))
              ->against_or(Storm_Query_Terms::not(['admin'])))
      ->fetchAll();

    $this->assertCount(0, $result);
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`level`) AGAINST(\'-admin\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withOrClause_LevelInviteOrFooLikePremierShouldAnswersUser_AdminAndUser_Invite() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->or((new Storm_Query_Criteria)
           ->eq('level', 'invite')
           ->like('foo', '%premier%'));

    $this->assertEquals(['user_admin', 'user_invite'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE ((`users`.`level` = \'invite\' OR `users`.`foo` LIKE \'%premier%\'))');
  }


  /** @test */
  public function withAndClause_LevelInviteOrFooLikePremierShouldAnswersOnlyUser_Invite() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->and((new Storm_Query_Criteria)
            ->eq('level', 'invite')
            ->like('foo', '%premier%'));

    $this->assertEquals(['user_invite'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE ((`users`.`level` = \'invite\' AND `users`.`foo` LIKE \'%premier%\'))');
  }


  /** @test */
  public function urlLikeHttpsDoubleDotSlashSlashShouldAnswersOnlyUserWithSlashes() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->like('url', '%https://my-server.org%');

    $this->assertEquals(['user_invite'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`url` LIKE \'%https://my-server.org%\')');
  }


  /** @test */
  public function websiteLikeHttpsDoubleDotSlashSlashPlusQuotePommeShouldAnswersOnlyUserWithSlashesPlusAndQuote() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->like('website', '%https://my-server.org/pomme+d\'api%');

    $this->assertEquals(['user_invite'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`website` LIKE \'%https://my-server.org/pomme+d\'api%\')');
  }


  /** @test */
  public function urlMatchHttpsDoubleDotSlashSlashShouldAnswersOnlyUserWithSlashes() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->match((new Storm_Query_MatchBoolean('website'))
              ->against_or('https://my-server.org/pomme+d\'api.php'));

    $this->assertEquals(['user_invite'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`website`) AGAINST(\'https://my-server.org/pomme+d\'api.php\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withIsNullPrefsShouldAnswersUser_100_Only() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->is_null('prefs');

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(100)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`prefs` IS null)');
  }


  /** @test */
  public function withIsNotNullPrefsShouldAnswersUsers_101_and_102() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->not_is_null('prefs');

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(101),
                         Storm_Test_LoaderQueryUser::find(102)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`prefs` IS NOT null)');
  }


  /** @test */
  public function withNotInInviteAdministrateurShouldAnswerUser_100_Only() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->not_in('level', ['invite', 'administrateur']);

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(100)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`level` NOT IN (\'invite\', \'administrateur\'))');
  }


  /** @test */
  public function withClauseLimitOneInIntAndStringShouldReturnOnlyUser_100() {
    $this->assertEquals([Storm_Test_LoaderQueryUser::find(100)],
                        Storm_Test_LoaderQueryUser::query()
                        ->limit(1)
                        ->fetchAll());

    $this->assertSql('SELECT `users`.* FROM `users` LIMIT 1');
  }


  /** @test */
  public function withEqualLengthOnColumnLevelShouldAnswerUser_102_Only() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->eq(Storm_Query_Key::length('level'), 6);

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(102)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (LENGTH(`users`.`level`) = 6)');
  }


  /** @test */
  public function withNotEqualLengthOnColumnLevelShouldAnswerUser_100_101() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->not_eq(Storm_Query_Key::length('level'), 6);

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(100),
                         Storm_Test_LoaderQueryUser::find(101)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (LENGTH(`users`.`level`) != 6)');
  }


  /** @test */
  public function withGreaterLengthOnColumnLevelShouldAnswerUser_100_Only() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->gt(Storm_Query_Key::length('level'), 14);

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(100)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (LENGTH(`users`.`level`) > 14)');
  }


  /** @test */
  public function withGreaterEqualLengthOnColumnLevelShouldAnswerUser_100_101() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->gt_eq(Storm_Query_Key::length('level'), 14);

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(100),
                         Storm_Test_LoaderQueryUser::find(101)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (LENGTH(`users`.`level`) >= 14)');
  }


  /** @test */
  public function withLesserLengthOnColumnLevelShouldAnswerUser_102_Only() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->lt(Storm_Query_Key::length('level'), 14);

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(102)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (LENGTH(`users`.`level`) < 14)');
  }


  /** @test */
  public function withLesserEqualLengthOnColumnLevelShouldAnswerUser_101_102() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->lt_eq(Storm_Query_Key::length('level'), 14);

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(101),
                         Storm_Test_LoaderQueryUser::find(102)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (LENGTH(`users`.`level`) <= 14)');
  }


  /** @test */
  public function withEqualLeftWidth4OnColumnDateShouldAnswerUser_101_Only() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->eq(Storm_Query_Key::left('date', 4), '2020');

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(101)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (LEFT(`users`.`date`, 4) = \'2020\')');
  }


  /** @test */
  public function withNotEqualLeftWidth4OnColumnDateShouldAnswerUser_100_102() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->not_eq(Storm_Query_Key::left('date', 4), '2020');

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(100),
                         Storm_Test_LoaderQueryUser::find(102)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (LEFT(`users`.`date`, 4) != \'2020\')');
  }


  /** @test */
  public function withGreaterLeftWidth4OnColumnDateShouldAnswerUser_100_Only() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->gt(Storm_Query_Key::left('date', 4), '2021');

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(100)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (LEFT(`users`.`date`, 4) > \'2021\')');
  }


  /** @test */
  public function withGreaterEqualLeftColumnDateShouldAnswerUser_100_102() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->gt_eq(Storm_Query_Key::left('date', 4), '2021');

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(100),
                         Storm_Test_LoaderQueryUser::find(102)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (LEFT(`users`.`date`, 4) >= \'2021\')');
  }


  /** @test */
  public function withLesserLeftColumnDateShouldAnswerUser_101_Only() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->lt(Storm_Query_Key::left('date', 4), '2021');

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(101)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (LEFT(`users`.`date`, 4) < \'2021\')');
  }


  /** @test */
  public function withLesserEqualLeftColumnDateShouldAnswerUser_101_102() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->lt_eq(Storm_Query_Key::left('date', 4), '2021');

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(101),
                         Storm_Test_LoaderQueryUser::find(102)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (LEFT(`users`.`date`, 4) <= \'2021\')');
  }


  /** @test */
  public function withClauseLimitPageShouldPaginate() {
    // LIMIT 0
    $this->assertEquals([],
                        Storm_Test_LoaderQueryUser::query()
                        ->limit_page(0, 0)
                        ->fetchAll());
    // LIMIT 1
    $this->assertEquals([Storm_Test_LoaderQueryUser::find(100)],
                        Storm_Test_LoaderQueryUser::query()
                        ->limit_page(0, 1)
                        ->fetchAll());
    // LIMIT 0
    $this->assertEquals([],
                        Storm_Test_LoaderQueryUser::query()
                        ->limit_page(1, 0)
                        ->fetchAll());
    // LIMIT 2
    $this->assertEquals([Storm_Test_LoaderQueryUser::find(100),
                         Storm_Test_LoaderQueryUser::find(101)],
                        Storm_Test_LoaderQueryUser::query()
                        ->limit_page(0, 2)
                        ->fetchAll());
    // LIMIT 2
    $this->assertEquals([Storm_Test_LoaderQueryUser::find(100),
                         Storm_Test_LoaderQueryUser::find(101)],
                        Storm_Test_LoaderQueryUser::query()
                        ->limit_page(1, 2)
                        ->fetchAll());
    // LIMIT 2 OFFSET 2
    $this->assertEquals([Storm_Test_LoaderQueryUser::find(102)],
                        Storm_Test_LoaderQueryUser::query()
                        ->limit_page(2, 2)
                        ->fetchAll());
    // LIMIT 5 OFFSET 10
    $this->assertEquals([],
                        Storm_Test_LoaderQueryUser::query()
                        ->limit_page(3, 5)
                        ->fetchAll());

    $this->assertSqlCount('SELECT `users`.* FROM `users` LIMIT 2', 2);
    $this->assertSqlCount('SELECT `users`.* FROM `users` LIMIT 0', 2);
    $this->assertSqlCount('SELECT `users`.* FROM `users` LIMIT 1', 1);
    $this->assertSqlCount('SELECT `users`.* FROM `users` LIMIT 2 OFFSET 2', 1);
    $this->assertSqlCount('SELECT `users`.* FROM `users` LIMIT 5 OFFSET 10', 1);
  }


  /** @test */
  public function withClauseLimitPageAndAttributeShouldWork() {
    $this->assertEquals([Storm_Test_LoaderQueryUser::find(100),
                         Storm_Test_LoaderQueryUser::find(102)],
                        Storm_Test_LoaderQueryUser::query()
                        ->start('foo', 'premier')
                        ->limit_page(1, 2)
                        ->fetchAll());

    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`foo` LIKE \'premier%\') LIMIT 2');
  }


  /** @test */
  public function queriesObserverShouldDisplaySqlAndCallCount() {
    foreach (range(1, 2) as $range)
      Storm_Test_LoaderQueryUser::query()
        ->eq('id', 100)
        ->fetchAll();

    foreach (range(1, 5) as $range)
      Storm_Test_LoaderQueryUser::query()
        ->gt('id', 100)
        ->fetchAll();

    Storm_Test_LoaderQueryUser::query()
      ->eq('id', 100)
      ->fetchAll();

    foreach (range(1, 4) as $range)
      Storm_Test_LoaderQueryUser::query()
        ->in('id', [100, 102])
        ->gt('id', 100)
        ->fetchAll();

    $this->assertSqlCount('SELECT `users`.* FROM `users` WHERE (`users`.`id` = 100)',
                          3);
    $this->assertSqlCount('SELECT `users`.* FROM `users` WHERE (`users`.`id` > 100)',
                          5);
    $this->assertSqlCount('SELECT `users`.* FROM `users` WHERE (`users`.`id` IN (100, 102) AND `users`.`id` > 100)',
                          4);
  }


  /** @test */
  public function whereMultipleCallShouldAssemble_2_Queries() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->gt('id', 100);

    $this->assertEquals(['user_administrateur', 'user_invite'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());

    $query->in('id', [101]);

    $this->assertEquals(['user_administrateur'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());

    $this->assertSqlEquals(['SELECT `users`.* FROM `users` WHERE (`users`.`id` > 100)',
                            'SELECT `users`.* FROM `users` WHERE (`users`.`id` > 100 AND `users`.`id` IN (101))']);
  }
}




class Storm_Test_LoaderQueryVolatileClauseOrderTest
  extends Storm_Test_ModelTestCase {

  protected array $_storm_scopes = ['LoaderQueryVolatileClauseOrderTest'];

  public function setUp(): void
  {
    parent::setUp();

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 100,
                    'login' => 'user_admin',
                    'level' => 'invite admin redacteur',
                    'foo' => 'first second third',
                    'order1' => 'abcd',
                    'order2' => 2,
                    'order3' => 'defg',
                    'order4' => 'weburl',
                    'order5' => '234']);

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 101,
                    'login' => 'user_administrateur',
                    'level' => 'administrateur',
                    'foo' => 'third second forth',
                    'order1' => 'abcdef',
                    'order2' => 3,
                    'order3' => 'defg',
                    'order4' => 'no',
                    'order5' => '12']);

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 102,
                    'login' => 'user_invite',
                    'level' => 'invite',
                    'foo' => 'forth fifth seven',
                    'order1' => 'cdef',
                    'order2' => 1,
                    'order3' => 'klmn',
                    'order4' => '',
                    'order5' => '111']);
  }


  /** @test */
  public function with_Order1Desc_ShouldAnswersInOrdered() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->order_desc('order1');

    $this->assertEquals(['user_invite', 'user_administrateur', 'user_admin'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` ORDER BY `users`.`order1` DESC');
  }


  /** @test */
  public function with_Order3Asc_Order2Desc_ShouldAnswersOrderedUser() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->order('order3')
      ->order_desc('order2');

    $this->assertEquals(['user_administrateur', 'user_admin', 'user_invite'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` ORDER BY `users`.`order3` ASC, `users`.`order2` DESC');
  }


  /** @test */
  public function with_Order3Desc_Order2Desc_ShouldAnswersOrderedUser() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->order_desc('order3')
      ->order_desc('order2');

    $this->assertEquals(['user_invite', 'user_administrateur', 'user_admin'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` ORDER BY `users`.`order3` DESC, `users`.`order2` DESC');
  }


  /** @test */
  public function withMatch_Level_Administrateur_ShouldAnswersOrderedUser() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->order((new Storm_Query_MatchRating('level'))
              ->against_or('administrateur'));

    $this->assertEquals(['user_admin', 'user_invite', 'user_administrateur'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` ORDER BY MATCH(`users`.`level`) AGAINST(\'administrateur\') ASC');
  }


  /** @test */
  public function withMatch_Level_Administrateur_Redacteur_Desc_ShouldAnswersOrderedUser() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->order_desc((new Storm_Query_MatchRating('level'))
                   ->against_or('administrateur redacteur'));

    $this->assertEquals(['user_administrateur', 'user_admin', 'user_invite'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` ORDER BY MATCH(`users`.`level`) AGAINST(\'administrateur redacteur\') DESC');
  }


  /** @test */
  public function withMatchInBooleanMode_Level_Administrateur_Redacteur_Desc_ShouldAnswersOrderedUser() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->order_desc((new Storm_Query_MatchBoolean('level'))
                   ->against_or('administrateur redacteur'));

    $this->assertEquals(['user_admin', 'user_administrateur', 'user_invite'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` ORDER BY MATCH(`users`.`level`) AGAINST(\'administrateur redacteur\' IN BOOLEAN MODE) DESC');
  }


  /** @test */
  public function withMatch_FooDescExacteInBooleanMode_Forth_Fifth_Level_Administrateur_Redacteur_ShouldAnswersOrderedUser() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->order_desc((new Storm_Query_MatchBoolean('foo'))
                   ->exact_or('forth fifth'))
      ->order((new Storm_Query_MatchRating('level'))
              ->against_or('administrateur redacteur'));

    $this->assertEquals(['user_invite', 'user_admin', 'user_administrateur'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` ORDER BY MATCH(`users`.`foo`) AGAINST(\'"forth fifth"\' IN BOOLEAN MODE) DESC, MATCH(`users`.`level`) AGAINST(\'administrateur redacteur\') ASC');
  }


  /** @test */
  public function withMatch_FooLevelStrict_InviteRedacteur_SecondThird_ShouldAnswersOrderedUser() {
    // https://www.php.net/manual/fr/function.usort
    // If two members compare as equal, they retain their original order. Prior to PHP 8.0.0, their relative order in the sorted array was undefined.
    if (version_compare(PHP_VERSION, '8.0.0', '<'))
      return;

    $query = Storm_Test_LoaderQueryUser::query()
      ->order((new Storm_Query_MatchBoolean('foo,level'))
              ->against_and('invite redacteur')
              ->against_and('second third'));

    $this->assertEquals([101, 102, 100],
                        array_map(fn($row) => $row->getId(), $query->fetchAll()));
    $this->assertSql('SELECT `users`.* FROM `users` ORDER BY MATCH(`users`.`foo`, `users`.`level`) AGAINST(\'+(invite redacteur) +(second third)\' IN BOOLEAN MODE) ASC');
  }


  /** @test */
  public function selectUsersGroupByOrder3ShouldAnswersUserAdminAndUserInvite() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->group('order3');

    $this->assertEquals(['user_admin', 'user_invite'],
                        (new Storm_Model_Collection($query->fetchAll()))
                        ->collect('login')
                        ->getArrayCopy());
    $this->assertSql('SELECT `users`.* FROM `users` GROUP BY `users`.`order3`');
  }


  /** @test */
  public function withSelectShouldAnswersWithRowModel() {
    $results = Storm_Test_LoaderQueryUser::query()
      ->select(['login', 'level'])
      ->fetchAll();

    $user = Storm_Test_LoaderQueryUser::find(100);
    $this->assertEquals(['login' => $user->getLogin(),
                         'level' => $user->getLevel()],
                        $results[0]->toArray());
    $this->assertTrue($results[0] instanceof Storm_Model_Row);

    $user = Storm_Test_LoaderQueryUser::find(101);
    $this->assertEquals(['login' => $user->getLogin(),
                         'level' => $user->getLevel()],
                        $results[1]->toArray());
    $this->assertTrue($results[1] instanceof Storm_Model_Row);

    $user = Storm_Test_LoaderQueryUser::find(102);
    $this->assertEquals(['login' => $user->getLogin(),
                         'level' => $user->getLevel()],
                        $results[2]->toArray());
    $this->assertTrue($results[2] instanceof Storm_Model_Row);
    $this->assertSql('SELECT `users`.`login`, `users`.`level` FROM `users`');
  }


  /** @test */
  public function with_Order4WithNoValueDesc_ShouldAnswersOrderedUser_101_100_102() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->order_desc(Storm_Query_Clause::orderValue('order4', 'no'));

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(101),
                         Storm_Test_LoaderQueryUser::find(100),
                         Storm_Test_LoaderQueryUser::find(102)],
                        $query->fetchAll());
    $this->assertSqlCount('SELECT `users`.* FROM `users` ORDER BY `users`.`order4` = \'no\' DESC', 1);
  }


  /** @test */
  public function with_Order1_ModeField_ShouldAnswersOrderedUser_101_102_100() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->order(Storm_Query_Clause::orderField('order1', ['cdef', 'abcd']));

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(101),
                         Storm_Test_LoaderQueryUser::find(102),
                         Storm_Test_LoaderQueryUser::find(100)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` ORDER BY FIELD(`users`.`order1`, \'cdef\', \'abcd\') ASC');
  }


  /** @test */
  public function with_Order5_ModeCastInt_ShouldAnswersOrderedUser_101_102_100() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->order(Storm_Query_Clause::orderCast('order5'));

    $this->assertEquals([Storm_Test_LoaderQueryUser::find(101),
                         Storm_Test_LoaderQueryUser::find(102),
                         Storm_Test_LoaderQueryUser::find(100)],
                        $query->fetchAll());
    $this->assertSql('SELECT `users`.* FROM `users` ORDER BY CAST(`users`.`order5` AS SIGNED INTEGER) ASC');
  }
}




class Storm_Test_LoaderQueryGroupWithMinMaxUser extends Storm_Test_ModelTestCase {

  protected array $_storm_scopes = ['LoaderQueryGroupWithMinMaxUser'];

  public function setUp(): void
  {
    parent::setUp();

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 100,
                    'login' => 'user_admin',
                    'date' => '2022-01-01',
                    'count' => 102,
                    'group' => 'group1',
                    'field' => 'field1']);

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 101,
                    'login' => 'user_administrateur',
                    'date' => '2020-01-01',
                    'count' => 15000,
                    'group' => 'group1',
                    'field' => 'field1']);

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 102,
                    'login' => 'user_invite',
                    'date' => '2025-01-06',
                    'count' => 2401,
                    'group' => 'group2',
                    'field' => 'field2']);

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 103,
                    'login' => 'user_bib',
                    'date' => '2000-12-01',
                    'count' => 98312,
                    'group' => 'group2',
                    'field' => 'field3']);

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 104,
                    'login' => 'user_portail',
                    'date' => '2021-01-01',
                    'count' => 1000,
                    'group' => 'group1',
                    'field' => 'field4']);
  }


  /** @test */
  public function groupByGroupAndSelectMaxDateWithHavingMaxDateShouldAnswers_100() {
    $row_results = Storm_Test_LoaderQueryUser::query()
      ->select(['id', 'group', 'max' => Storm_Query_Key::max('date')])
      ->group('group')
      ->having_gt('max', '2018-01-01')
      ->having_lt('max', '2024-01-01')
      ->fetchAll();

    $this->assertCount(1, $row_results);
    $this->assertEquals([['id' => 100,
                          'group' => 'group1',
                          'max' => '2022-01-01']],
                        array_map(fn($row) => $row->toArray(),
                                  $row_results));
    $this->assertSql("SELECT `users`.`id`, `users`.`group`, MAX(`users`.`date`) AS `max` FROM `users` GROUP BY `users`.`group` HAVING (`max` > '2018-01-01') AND (`max` < '2024-01-01')");
  }


  /** @test */
  public function groupByGroupAndFieldShouldAnswersUserId_100_102_103_104() {
    $row_results = Storm_Test_LoaderQueryUser::query()
      ->select(['id', 'group', 'field'])
      ->group('group')
      ->group('field')
      ->fetchAll();

    $this->assertCount(4, $row_results);
    $this->assertEquals([['id' => 100,
                          'group' => 'group1',
                          'field' => 'field1'],
                         ['id' => 102,
                          'group' => 'group2',
                          'field' => 'field2'],
                         ['id' => 103,
                          'group' => 'group2',
                          'field' => 'field3'],
                         ['id' => 104,
                          'group' => 'group1',
                          'field' => 'field4']],
                        array_map(fn($row) => $row->toArray(),
                                  $row_results));
    $this->assertSql('SELECT `users`.`id`, `users`.`group`, `users`.`field` FROM `users` GROUP BY `users`.`group`, `users`.`field`');
  }


  /** @test */
  public function selectUsersGroupByGroupWithSelectMinOnDateShouldAnswersUserAdministrateurAndUserBib() {
    $row_results = Storm_Test_LoaderQueryUser::query()
      ->select([Storm_Query_Key::min('date'), 'login', 'group'])
      ->group('group')
      ->fetchAll();

    $this->assertCount(2, $row_results);
    $this->assertEquals([['date' => '2020-01-01',
                          'login' => 'user_administrateur',
                          'group' => 'group1'],
                         ['date' => '2000-12-01',
                          'login' => 'user_bib',
                          'group' => 'group2']],
                        array_map(fn($row) => $row->toArray(),
                                  $row_results));
    $this->assertSql('SELECT MIN(`users`.`date`), `users`.`login`, `users`.`group` FROM `users` GROUP BY `users`.`group`');
  }


  /** @test */
  public function selectUsersGroupByGroupWithSelectMaxOnCountShouldAnswersUserAdministrateurAndUserBib() {
    $row_results = Storm_Test_LoaderQueryUser::query()
      ->select([Storm_Query_Key::max('count'), 'login', 'group'])
      ->group('group')
      ->fetchAll();

    $this->assertCount(2, $row_results);
    $this->assertEquals([['count' => 15000,
                          'login' => 'user_administrateur',
                          'group' => 'group1'],
                         ['count' => 98312,
                          'login' => 'user_bib',
                          'group' => 'group2']],
                        array_map(fn($row) => $row->toArray(),
                                  $row_results));
    $this->assertSql('SELECT MAX(`users`.`count`), `users`.`login`, `users`.`group` FROM `users` GROUP BY `users`.`group`');
  }


  /** @test */
  public function selectUsersGroupByGroupWithSelectSumOnCountShouldAnswersUserAdminAndUserInvite() {
    $row_results = Storm_Test_LoaderQueryUser::query()
      ->select([Storm_Query_Key::sum('count'), 'login', 'group'])
      ->group('group')
      ->fetchAll();

    $this->assertCount(2, $row_results);
    $this->assertEquals([['count' => 16102,
                          'login' => 'user_admin',
                          'group' => 'group1'],
                         ['count' => 100713,
                          'login' => 'user_invite',
                          'group' => 'group2']],
                        array_map(fn($row) => $row->toArray(),
                                  $row_results));
    $this->assertSql('SELECT SUM(`users`.`count`), `users`.`login`, `users`.`group` FROM `users` GROUP BY `users`.`group`');
  }


  /** @test */
  public function selectUsersGroupByGroupWithSelectCountShouldAnswersUserAdminAndUserInvite() {
    $row_results = Storm_Test_LoaderQueryUser::query()
      ->select(['count_users' => Storm_Query_Key::count(), 'login', 'group'])
      ->group('group')
      ->fetchAll();

    $this->assertCount(2, $row_results);
    $this->assertEquals([['count_users' => 3,
                          'login' => 'user_admin',
                          'group' => 'group1'],
                         ['count_users' => 2,
                          'login' => 'user_invite',
                          'group' => 'group2']],
                        array_map(fn($row) => $row->toArray(),
                                  $row_results));
    $this->assertSql('SELECT COUNT(*) AS `count_users`, `users`.`login`, `users`.`group` FROM `users` GROUP BY `users`.`group`');
  }


  /** @test */
  public function groupByGroupWithSelectCountAndHaving_GT_2_ShouldAnswersUserAdmin() {
    $row_results = Storm_Test_LoaderQueryUser::query()
      ->select([Storm_Query_Key::count(), 'login', 'group'])
      ->group('group')
      ->having_gt('count_by', 2)
      ->fetchAll();

    $this->assertCount(1, $row_results);
    $this->assertEquals([['count_by' => 3,
                          'login' => 'user_admin',
                          'group' => 'group1']],
                        array_map(fn($row) => $row->toArray(),
                                  $row_results));
    $this->assertSql('SELECT COUNT(*) AS `count_by`, `users`.`login`, `users`.`group` FROM `users` GROUP BY `users`.`group` HAVING (`count_by` > 2)');
  }


  /** @test */
  public function groupByGroupWithSelectCountAndHaving_GT_EQ_3_ShouldAnswersUserAdmin() {
    $row_results = Storm_Test_LoaderQueryUser::query()
      ->select([Storm_Query_Key::count(), 'login', 'group'])
      ->group('group')
      ->having_gt_eq('count_by', 3)
      ->fetchAll();

    $this->assertCount(1, $row_results);
    $this->assertEquals([['count_by' => 3,
                          'login' => 'user_admin',
                          'group' => 'group1']],
                        array_map(fn($row) => $row->toArray(),
                                  $row_results));
    $this->assertSql('SELECT COUNT(*) AS `count_by`, `users`.`login`, `users`.`group` FROM `users` GROUP BY `users`.`group` HAVING (`count_by` >= 3)');
  }


  /** @test */
  public function groupByGroupWithSelectCountAndHaving_EQ_3_ShouldAnswersUserAdmin() {
    $row_results = Storm_Test_LoaderQueryUser::query()
      ->select([Storm_Query_Key::count(), 'login', 'group'])
      ->group('group')
      ->having_eq('count_by', 3)
      ->fetchAll();

    $this->assertCount(1, $row_results);
    $this->assertEquals([['count_by' => 3,
                          'login' => 'user_admin',
                          'group' => 'group1']],
                        array_map(fn($row) => $row->toArray(),
                                  $row_results));
    $this->assertSql('SELECT COUNT(*) AS `count_by`, `users`.`login`, `users`.`group` FROM `users` GROUP BY `users`.`group` HAVING (`count_by` = 3)');
  }


  /** @test */
  public function groupByGroupWithSelectCountAndHaving_NOT_EQ_2_ShouldAnswersUserAdmin() {
    $row_results = Storm_Test_LoaderQueryUser::query()
      ->select([Storm_Query_Key::count(), 'login', 'group'])
      ->group('group')
      ->having_not_eq('count_by', 2)
      ->fetchAll();

    $this->assertCount(1, $row_results);
    $this->assertEquals([['count_by' => 3,
                          'login' => 'user_admin',
                          'group' => 'group1']],
                        array_map(fn($row) => $row->toArray(),
                                  $row_results));
    $this->assertSql('SELECT COUNT(*) AS `count_by`, `users`.`login`, `users`.`group` FROM `users` GROUP BY `users`.`group` HAVING (`count_by` != 2)');
  }


  /** @test */
  public function groupByGroupWithSelectCountAndHaving_LT_3_ShouldAnswersUserInvite() {
    $row_results = Storm_Test_LoaderQueryUser::query()
      ->select([Storm_Query_Key::count(), 'login', 'group'])
      ->group('group')
      ->having_lt('count_by', 3)
      ->fetchAll();

    $this->assertCount(1, $row_results);
    $this->assertEquals([['count_by' => 2,
                          'login' => 'user_invite',
                          'group' => 'group2']],
                        array_map(fn($row) => $row->toArray(),
                                  $row_results));
    $this->assertSql('SELECT COUNT(*) AS `count_by`, `users`.`login`, `users`.`group` FROM `users` GROUP BY `users`.`group` HAVING (`count_by` < 3)');
  }


  /** @test */
  public function groupByGroupWithSelectCountAndHaving_LT_EQ_2_ShouldAnswersUserInvite() {
    $row_results = Storm_Test_LoaderQueryUser::query()
      ->select([Storm_Query_Key::count(), 'login', 'group'])
      ->group('group')
      ->having_lt_eq('count_by', 2)
      ->fetchAll();

    $this->assertCount(1, $row_results);
    $this->assertEquals([['count_by' => 2,
                          'login' => 'user_invite',
                          'group' => 'group2']],
                        array_map(fn($row) => $row->toArray(),
                                  $row_results));
    $this->assertSql('SELECT COUNT(*) AS `count_by`, `users`.`login`, `users`.`group` FROM `users` GROUP BY `users`.`group` HAVING (`count_by` <= 2)');
  }
}




class Storm_Test_LoaderQueryVolatileDistinctTest
  extends Storm_Test_ModelTestCase {

  protected array $_storm_scopes = ['LoaderQueryVolatileDistinctTest'];

  public function setUp(): void
  {
    parent::setUp();

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 100,
                    'distinct1' => 'abcd',
                    'distinct2' => 1234]);

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 101,
                    'distinct1' => 'abcd',
                    'distinct2' => 1234]);

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 102,
                    'distinct1' => 'abcd',
                    'distinct2' => 4567]);

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 103,
                    'distinct1' => 'zyw',
                    'distinct2' => 987]);

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 104,
                    'distinct1' => 'fed',
                    'distinct2' => 987]);

    $this->fixture(Storm_Test_LoaderQueryActor::class,
                   ['id' => 200,
                    'name' => 'totoro',
                    'user_id' => 100]);

    $this->fixture(Storm_Test_LoaderQueryActor::class,
                   ['id' => 201,
                    'name' => 'totoro',
                    'user_id' => 102]);

    $this->fixture(Storm_Test_LoaderQueryActor::class,
                   ['id' => 202,
                    'name' => 'chichiro',
                    'user_id' => 102]);
  }


  /** @test */
  public function withDistinct1AndDistinct2ShouldAnswers4Result() {
    $row_results = Storm_Test_LoaderQueryUser::query()
      ->distinct(['distinct1', 'distinct2'])
      ->fetchAll();

    $this->assertEquals([['distinct1' => 'abcd', 'distinct2' => 1234],
                         ['distinct1' => 'abcd', 'distinct2' => 4567],
                         ['distinct1' => 'zyw', 'distinct2' => 987],
                         ['distinct1' => 'fed', 'distinct2' => 987]],
                        array_map(fn($row) => $row->toArray(),
                                  $row_results));
    $this->assertSql('SELECT DISTINCT `users`.`distinct1`, `users`.`distinct2` FROM `users`');
  }


  /** @test */
  public function withInUserIdShouldAnswers_2_Users_100_102() {
    $row_results = Storm_Test_LoaderQueryUser::query('tab1')
      ->select('id')
      ->in('id', Storm_Test_LoaderQueryActor::subIn('tab2')
           ->distinct('user_id'))
      ->fetchAll();

    $this->assertEquals([['id' => 100],
                         ['id' => 102]],
                        array_map(fn($row) => $row->toArray(),
                                  $row_results));
    $this->assertSql('SELECT `tab1`.`id` FROM `users` AS `tab1` WHERE (`tab1`.`id` IN (SELECT DISTINCT `tab2`.`user_id` FROM `actors` AS `tab2`))');
  }
}




class Storm_Test_LoaderQueryVolatileJoinInnerTest
  extends Storm_Test_ModelTestCase {

  protected array $_storm_scopes = ['LoaderQueryVolatileJoinInnerTest'];

  public function setUp(): void
  {
    parent::setUp();

    $this->fixture(Storm_Test_LoaderQueryMovie::class,
                   ['id' => 1,
                    'name' => 'Movie name 1',
                    'date' => '2020']);
    $this->fixture(Storm_Test_LoaderQueryMovie::class,
                   ['id' => 2,
                    'name' => 'Movie name 2',
                    'date' => '2022']);
    $this->fixture(Storm_Test_LoaderQueryMovie::class,
                   ['id' => 3,
                    'name' => 'Movie name 3',
                    'date' => '2021']);
    $this->fixture(Storm_Test_LoaderQueryMovie::class,
                   ['id' => 4,
                    'name' => 'Movie name 4',
                    'date' => '2018']);
    $this->fixture(Storm_Test_LoaderQueryMovie::class,
                   ['id' => 5,
                    'name' => 'Movie name 5',
                    'date' => '2019']);

    $this->fixture(Storm_Test_LoaderQueryActor::class,
                   ['id' => 1,
                    'movie_id' => 3,
                    'name' => 'Actor name 1',
                    'age' => 25]);
    $this->fixture(Storm_Test_LoaderQueryActor::class,
                   ['id' => 2,
                    'movie_id' => 1,
                    'name' => 'Actor name 2',
                    'age' => 52]);
    $this->fixture(Storm_Test_LoaderQueryActor::class,
                   ['id' => 3,
                    'movie_id' => 4,
                    'name' => 'Actor name 3',
                    'age' => 36]);
    $this->fixture(Storm_Test_LoaderQueryActor::class,
                   ['id' => 4,
                    'movie_id' => 3,
                    'name' => 'Actor name 4',
                    'age' => 20]);
  }


  /** @test */
  public function joinWithOrConditionOnParentAndChildShouldReturn_Id_3_3_4() {
    $row_results = Storm_Test_LoaderQueryMovie::join('tab1')
      ->all_select()
      ->inner(Storm_Test_LoaderQueryActor::inner('tab2')
              ->on_eq('id', 'movie_id'))
      ->or_sub(['tab1' => (new Storm_Query_Criteria)
                ->eq('date', '2021'),
                'tab2' => (new Storm_Query_Criteria)
                ->in('age', [36])])
      ->fetchAll();

    $this->assertEquals([['id' => 3,
                          'name' => 'Movie name 3',
                          'date' => '2021'],
                         ['id' => 3,
                          'name' => 'Movie name 3',
                          'date' => '2021'],
                         ['id' => 4,
                          'name' => 'Movie name 4',
                          'date' => '2018']],
                        array_map(fn($row) => $row->attributesToArray(),
                                  $row_results));

    $this->assertSql('SELECT `tab1`.* FROM `movies` AS `tab1` INNER JOIN `actors` AS `tab2` ON `tab1`.`id` = `tab2`.`movie_id` WHERE ((`tab1`.`date` = \'2021\' OR `tab2`.`age` IN (36)))');
  }


  /** @test */
  public function existsWithAndConditionOnParentAndChildShouldReturn_Id_3() {
    $row_results = Storm_Test_LoaderQueryMovie::query('tab1')
      ->exists(Storm_Test_LoaderQueryActor::subExists('tab2')
               ->on_eq('id', 'movie_id'))
      ->and_sub(['tab1' => (new Storm_Query_Criteria)
                 ->eq('date', '2021'),
                 'tab2' => (new Storm_Query_Criteria)
                 ->in('age', [25, 36])])
      ->fetchAll();

    $this->assertEquals([['id' => 3,
                          'name' => 'Movie name 3',
                          'date' => '2021']],
                        array_map(fn($row) => $row->attributesToArray(),
                                  $row_results));

    $this->assertSql('SELECT `tab1`.* FROM `movies` AS `tab1` WHERE (EXISTS (SELECT \'x\' FROM `actors` AS `tab2` WHERE (`tab1`.`id` = `tab2`.`movie_id`)) AND (`tab1`.`date` = \'2021\' AND `tab2`.`age` IN (25, 36)))');
  }


  /** @test */
  public function joinWithAnd_OrConditionOnParentAndChildShouldReturn_Id_3() {
    $row_results = Storm_Test_LoaderQueryMovie::join('tab1')
      ->select('name')
      ->inner(Storm_Test_LoaderQueryActor::inner('tab2')
              ->on_eq('id', 'movie_id')
              ->select('name'))
      ->and_sub(['tab1' => (new Storm_Query_Criteria)
                 ->eq('date', '2021'),
                 'tab2' => (new Storm_Query_Criteria)
                 ->in('age', [20, 36])])
      ->or_sub(['tab1' => (new Storm_Query_Criteria)
                ->eq('name', 'Movie name 3'),
                'tab2' => (new Storm_Query_Criteria)
                ->not_end('name', 'name 1')])
      ->fetchAll();

    $this->assertEquals([['tab1_name' => 'Movie name 3',
                          'tab2_name' => 'Actor name 4']],
                        array_map(fn($row) => $row->attributesToArray(),
                                  $row_results));

    $this->assertSql('SELECT `tab1`.`name` AS `tab1_name`, `tab2`.`name` AS `tab2_name` FROM `movies` AS `tab1` INNER JOIN `actors` AS `tab2` ON `tab1`.`id` = `tab2`.`movie_id` WHERE ((`tab1`.`date` = \'2021\' AND `tab2`.`age` IN (20, 36)) AND (`tab1`.`name` = \'Movie name 3\' OR `tab2`.`name` NOT LIKE \'%name 1\'))');
  }


  /** @test */
  public function withInnerJoinShouldReturnIntersectMoviesActorsOnlyOneResult() {
    $results = Storm_Test_LoaderQueryMovie::join('tab1')
      ->select(['name', 'date'])
      ->inner(Storm_Test_LoaderQueryActor::inner('tab2')
              ->select(['name', 'age'])
              ->on_eq('id', 'movie_id')
              ->gt('age', 30))
      ->in('date', ['2019', '2020', '2021'])
      ->fetchAll();

    $row = $results[0];
    $this->assertCount(1, $results);
    $this->assertEquals('Movie name 1', $row->getTab1Name());
    $this->assertEquals('2020', $row->tab1_date);
    $this->assertEquals('Actor name 2', $row->getTab2Name());
    $this->assertEquals(52, $row->tab2_age);
    $this->assertTrue($row instanceof Storm_Model_Row);

    $this->assertSql('SELECT `tab1`.`name` AS `tab1_name`, `tab1`.`date` AS `tab1_date`, `tab2`.`name` AS `tab2_name`, `tab2`.`age` AS `tab2_age` FROM `movies` AS `tab1` INNER JOIN `actors` AS `tab2` ON `tab1`.`id` = `tab2`.`movie_id` AND `tab2`.`age` > 30 WHERE (`tab1`.`date` IN (\'2019\', \'2020\', \'2021\'))');
  }


  /** @test */
  public function withInnerJoinShouldReturnIntersectDedupMovieAndTwoActors() {
    $results = Storm_Test_LoaderQueryMovie::join('tab1')
      ->select(['name', 'date'])
      ->inner(Storm_Test_LoaderQueryActor::inner('tab2')
              ->select(['name', 'age'])
              ->on_eq('id', 'movie_id')
              ->lt('age', 30))
      ->in('date', ['2019', '2020', '2021'])
      ->fetchAll();

    $this->assertCount(2, $results);
    $this->assertEquals([['tab1_name' => 'Movie name 3',
                          'tab1_date' => '2021',
                          'tab2_name' => 'Actor name 1',
                          'tab2_age' => 25],
                         ['tab1_name' => 'Movie name 3',
                          'tab1_date' => '2021',
                          'tab2_name' => 'Actor name 4',
                          'tab2_age' => 20]],
                        array_map(fn($row) => $row->toArray(),
                                  $results));
    $this->assertTrue($results[0] instanceof Storm_Model_Row);

    $this->assertSql('SELECT `tab1`.`name` AS `tab1_name`, `tab1`.`date` AS `tab1_date`, `tab2`.`name` AS `tab2_name`, `tab2`.`age` AS `tab2_age` FROM `movies` AS `tab1` INNER JOIN `actors` AS `tab2` ON `tab1`.`id` = `tab2`.`movie_id` AND `tab2`.`age` < 30 WHERE (`tab1`.`date` IN (\'2019\', \'2020\', \'2021\'))');
  }


  /** @test */
  public function withInnerJoinWithoutSelectShouldReturnMovies() {
    $results = Storm_Test_LoaderQueryMovie::join('tab1')
      ->all_select()
      ->inner(Storm_Test_LoaderQueryActor::inner('tab2')
              ->on_eq('id', 'movie_id')
              ->gt('age', 30))
      ->in('date', ['2019', '2020', '2021'])
      ->fetchAll();

    $this->assertCount(1, $results);
    $this->assertEquals(['id' => 1,
                         'name' => 'Movie name 1',
                         'date' => '2020'],
                        $results[0]->toArray());
    $this->assertTrue($results[0] instanceof Storm_Model_Row);

    $this->assertSql('SELECT `tab1`.* FROM `movies` AS `tab1` INNER JOIN `actors` AS `tab2` ON `tab1`.`id` = `tab2`.`movie_id` AND `tab2`.`age` > 30 WHERE (`tab1`.`date` IN (\'2019\', \'2020\', \'2021\'))');
  }


  /** @test */
  public function queryPromoteInnerJoinWithoutSelectShouldReturnMovies() {
    $results = Storm_Test_LoaderQueryMovie::query()
      ->promoteToJoin('tab1')
      ->inner(Storm_Test_LoaderQueryActor::inner('tab2')
              ->on_eq('id', 'movie_id')
              ->gt('age', 30))
      ->in('date', ['2019', '2020', '2021'])
      ->fetchAll();

    $this->assertCount(1, $results);
    $this->assertEquals(['id' => 1,
                         'name' => 'Movie name 1',
                         'date' => '2020'],
                        $results[0]->toArray());
    $this->assertTrue($results[0] instanceof Storm_Model_Row);

    $this->assertSql('SELECT `tab1`.* FROM `movies` AS `tab1` INNER JOIN `actors` AS `tab2` ON `tab1`.`id` = `tab2`.`movie_id` AND `tab2`.`age` > 30 WHERE (`tab1`.`date` IN (\'2019\', \'2020\', \'2021\'))');
  }


  /** @test */
  public function queryPromoteSubInnerShouldReturnMovies() {
    $query = Storm_Test_LoaderQueryActor::query()
      ->select('movie_id')
      ->gt('age', 30)
      ->limit(100)
      ->order('age');

    $results = Storm_Test_LoaderQueryMovie::join('tab1')
      ->all_select()
      ->inner($query->promoteToInner('tab2')
              ->on_eq('id', 'movie_id'))
      ->in('date', ['2019', '2020', '2021'])
      ->fetchAll();

    $this->assertCount(1, $results);
    $this->assertEquals(['id' => 1,
                         'name' => 'Movie name 1',
                         'date' => '2020'],
                        $results[0]->toArray());

    $this->assertSql('SELECT `tab1`.* FROM `movies` AS `tab1` INNER JOIN `actors` AS `tab2` ON `tab1`.`id` = `tab2`.`movie_id` AND `tab2`.`age` > 30 WHERE (`tab1`.`date` IN (\'2019\', \'2020\', \'2021\'))');
  }


  /** @test */
  public function withInnerJoinWithSelectTab2AllShouldReturnActors() {
    $results = Storm_Test_LoaderQueryMovie::join('tab1')
      ->inner(Storm_Test_LoaderQueryActor::inner('tab2')
              ->on_eq('id', 'movie_id')
              ->all_select()
              ->gt('age', 30))
      ->in('date', ['2019', '2020', '2021'])
      ->fetchAll();

    $this->assertCount(1, $results);
    $this->assertEquals(['id' => 2,
                         'movie_id' => 1,
                         'name' => 'Actor name 2',
                         'age' => 52],
                        $results[0]->toArray());
    $this->assertTrue($results[0] instanceof Storm_Model_Row);

    $this->assertSql('SELECT `tab2`.* FROM `movies` AS `tab1` INNER JOIN `actors` AS `tab2` ON `tab1`.`id` = `tab2`.`movie_id` AND `tab2`.`age` > 30 WHERE (`tab1`.`date` IN (\'2019\', \'2020\', \'2021\'))');
  }


  /** @test */
  public function withInnerJoinWithSelectNameTab2AndSelectAllTab1ShouldReturnActors() {
    $results = Storm_Test_LoaderQueryMovie::join('tab1')
      ->all_select()
      ->inner(Storm_Test_LoaderQueryActor::inner('tab2')
              ->on_eq('id', 'movie_id')
              ->select('name')
              ->gt('age', 30))
      ->in('date', ['2019', '2020', '2021'])
      ->fetchAll();

    $this->assertCount(1, $results);
    $this->assertEquals(['tab2_name' => 'Actor name 2',
                         'id' => 1,
                         'name' => 'Movie name 1',
                         'date' => '2020'],
                        $results[0]->toArray());
    $this->assertTrue($results[0] instanceof Storm_Model_Row);

    $this->assertSql('SELECT `tab1`.*, `tab2`.`name` AS `tab2_name` FROM `movies` AS `tab1` INNER JOIN `actors` AS `tab2` ON `tab1`.`id` = `tab2`.`movie_id` AND `tab2`.`age` > 30 WHERE (`tab1`.`date` IN (\'2019\', \'2020\', \'2021\'))');
  }
}




class Storm_Test_LoaderQueryVolatileJoinLeftTest
  extends Storm_Test_ModelTestCase {

  protected array $_storm_scopes = ['LoaderQueryVolatileJoinLeftTest'];

  public function setUp(): void
  {
    parent::setUp();

    $this->fixture(Storm_Test_LoaderQueryMovie::class,
                   ['id' => 1,
                    'name' => 'Movie name 1',
                    'date' => '2020']);
    $this->fixture(Storm_Test_LoaderQueryMovie::class,
                   ['id' => 2,
                    'name' => 'Movie name 2',
                    'date' => '2022']);
    $this->fixture(Storm_Test_LoaderQueryMovie::class,
                   ['id' => 3,
                    'name' => 'Movie name 3',
                    'date' => '2021']);
    $this->fixture(Storm_Test_LoaderQueryMovie::class,
                   ['id' => 4,
                    'name' => 'Movie name 4',
                    'date' => '2018']);

    $this->fixture(Storm_Test_LoaderQueryActor::class,
                   ['id' => 1,
                    'movie_id' => 3,
                    'name' => 'Actor name 1',
                    'age' => 25]);
    $this->fixture(Storm_Test_LoaderQueryActor::class,
                   ['id' => 2,
                    'movie_id' => 1,
                    'name' => 'Actor name 2',
                    'age' => 52]);
    $this->fixture(Storm_Test_LoaderQueryActor::class,
                   ['id' => 3,
                    'movie_id' => 3,
                    'name' => 'Actor name 3',
                    'age' => 20]);
  }


  /** @test */
  public function withLeftJoinShouldReturn() {
    $row_results = Storm_Test_LoaderQueryMovie::join('tab1')
      ->select(['name'])
      ->left(Storm_Test_LoaderQueryActor::left('tab2')
             ->select(['name'])
             ->on_eq('id', 'movie_id'))
      ->fetchAll();

    $this->assertCount(5, $row_results);
    $this->assertEquals([['tab1_name' => 'Movie name 1',
                          'tab2_name' => 'Actor name 2'],
                         ['tab1_name' => 'Movie name 2',
                          'tab2_name' => null],
                         ['tab1_name' => 'Movie name 3',
                          'tab2_name' => 'Actor name 1'],
                         ['tab1_name' => 'Movie name 3',
                          'tab2_name' => 'Actor name 3'],
                         ['tab1_name' => 'Movie name 4',
                          'tab2_name' => null]],
                        array_map(fn($row) => $row->toArray(),
                                  $row_results));
    $this->assertTrue($row_results[0] instanceof Storm_Model_Row);

    $this->assertSql('SELECT `tab1`.`name` AS `tab1_name`, `tab2`.`name` AS `tab2_name` FROM `movies` AS `tab1` LEFT JOIN `actors` AS `tab2` ON `tab1`.`id` = `tab2`.`movie_id`');
  }
}




class Storm_Test_LoaderQueryVolatileJoinRightTest
  extends Storm_Test_ModelTestCase {

  protected array $_storm_scopes = ['LoaderQueryVolatileJoinRightTest'];

  public function setUp(): void
  {
    parent::setUp();

    $this->fixture(Storm_Test_LoaderQueryMovie::class,
                   ['id' => 1,
                    'name' => 'Movie name 1',
                    'date' => '2020']);
    $this->fixture(Storm_Test_LoaderQueryMovie::class,
                   ['id' => 2,
                    'name' => 'Movie name 2',
                    'date' => '2022']);
    $this->fixture(Storm_Test_LoaderQueryMovie::class,
                   ['id' => 3,
                    'name' => 'Movie name 3',
                    'date' => '2021']);

    $this->fixture(Storm_Test_LoaderQueryActor::class,
                   ['id' => 1,
                    'movie_id' => 4,
                    'name' => 'Actor name 1',
                    'age' => 25]);
    $this->fixture(Storm_Test_LoaderQueryActor::class,
                   ['id' => 2,
                    'movie_id' => 1,
                    'name' => 'Actor name 2',
                    'age' => 52]);
    $this->fixture(Storm_Test_LoaderQueryActor::class,
                   ['id' => 3,
                    'movie_id' => 1,
                    'name' => 'Actor name 3',
                    'age' => 20]);
  }


  /** @test */
  public function withRightJoinShouldReturn() {
    $row_results = Storm_Test_LoaderQueryMovie::join('tab1')
      ->select(['name'])
      ->right(Storm_Test_LoaderQueryActor::right('tab2')
              ->select(['name'])
              ->on_eq('id', 'movie_id'))
      ->fetchAll();

    $this->assertCount(3, $row_results);
    $this->assertEquals([['tab1_name' => null,
                          'tab2_name' => 'Actor name 1'],
                         ['tab1_name' => 'Movie name 1',
                          'tab2_name' => 'Actor name 2'],
                         ['tab1_name' => 'Movie name 1',
                          'tab2_name' => 'Actor name 3']],
                        array_map(fn($row) => $row->toArray(),
                                  $row_results));
    $this->assertTrue($row_results[0] instanceof Storm_Model_Row);

    $this->assertSql('SELECT `tab1`.`name` AS `tab1_name`, `tab2`.`name` AS `tab2_name` FROM `movies` AS `tab1` RIGHT JOIN `actors` AS `tab2` ON `tab1`.`id` = `tab2`.`movie_id`');
  }
}




class Storm_Test_LoaderQueryVolatileJoinMultipleTest
  extends Storm_Test_ModelTestCase {

  protected array $_storm_scopes = ['LoaderQueryVolatileJoinMultipleTest'];

  public function setUp(): void
  {
    parent::setUp();

    $this->fixture(Storm_Test_LoaderQueryMovie::class,
                   ['id' => 1,
                    'name' => 'Movie name 1',
                    'date' => '2020']);
    $this->fixture(Storm_Test_LoaderQueryMovie::class,
                   ['id' => 2,
                    'name' => 'Movie name 2',
                    'date' => '2022']);
    $this->fixture(Storm_Test_LoaderQueryMovie::class,
                   ['id' => 3,
                    'name' => 'Movie name 3',
                    'date' => '2021']);
    $this->fixture(Storm_Test_LoaderQueryMovie::class,
                   ['id' => 4,
                    'name' => 'Movie name 4',
                    'date' => '2018']);
    $this->fixture(Storm_Test_LoaderQueryMovie::class,
                   ['id' => 5,
                    'name' => 'Movie name 5',
                    'date' => '2019']);

    $this->fixture(Storm_Test_LoaderQueryActor::class,
                   ['id' => 1,
                    'movie_id' => 3,
                    'name' => 'Actor name 1',
                    'age' => 25]);
    $this->fixture(Storm_Test_LoaderQueryActor::class,
                   ['id' => 2,
                    'movie_id' => 1,
                    'name' => 'Actor name 2',
                    'age' => 52]);
    $this->fixture(Storm_Test_LoaderQueryActor::class,
                   ['id' => 3,
                    'movie_id' => 4,
                    'name' => 'Actor name 3',
                    'age' => 36]);
    $this->fixture(Storm_Test_LoaderQueryActor::class,
                   ['id' => 4,
                    'movie_id' => 3,
                    'name' => 'Actor name 4',
                    'age' => 20]);

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 1,
                    'libelle' => 'User 1',
                    'movie_id' => 2]);
    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 2,
                    'libelle' => 'User 2',
                    'movie_id' => 3]);
    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 3,
                    'libelle' => 'User 3',
                    'movie_id' => 4]);

    $this->fixture(Storm_Test_LoaderQueryProductor::class,
                   ['id' => 1,
                    'label' => 'Productor 1',
                    'user_id' => 1,
                    'birthday' => '01/01/2015']);
    $this->fixture(Storm_Test_LoaderQueryProductor::class,
                   ['id' => 2,
                    'label' => 'Productor 2',
                    'user_id' => 1,
                    'birthday' => '02/02/2016']);
    $this->fixture(Storm_Test_LoaderQueryProductor::class,
                   ['id' => 3,
                    'label' => 'Productor 3',
                    'user_id' => 3,
                    'birthday' => '03/03/2017']);
    $this->fixture(Storm_Test_LoaderQueryProductor::class,
                   ['id' => 4,
                    'label' => 'Productor 4',
                    'user_id' => 3,
                    'birthday' => '04/04/2018']);
  }


  /** @test */
  public function withInnerJoinShouldReturnIntersectMoviesActorsOnlyOneResult() {
    $results = Storm_Test_LoaderQueryMovie::join('tab1')
      ->select('date')
      ->inner(Storm_Test_LoaderQueryActor::inner('tab2')
              ->on_eq('id', 'movie_id')
              ->select('name')
              ->gt('age', 30))
      ->inner(Storm_Test_LoaderQueryUser::inner('tab3')
              ->on_eq('id', 'movie_id')
              ->select('libelle')
              ->right(Storm_Test_LoaderQueryProductor::right('tab4')
                      ->on_eq('id', 'user_id')
                      ->select(['label', 'birthday'])))
      ->fetchAll();

    $this->assertCount(4, $results);
    $this->assertSql('SELECT `tab1`.`date` AS `tab1_date`, `tab2`.`name` AS `tab2_name`, `tab3`.`libelle` AS `tab3_libelle`, `tab4`.`label` AS `tab4_label`, `tab4`.`birthday` AS `tab4_birthday` FROM `movies` AS `tab1` INNER JOIN `actors` AS `tab2` ON `tab1`.`id` = `tab2`.`movie_id` AND `tab2`.`age` > 30 INNER JOIN `users` AS `tab3` ON `tab1`.`id` = `tab3`.`movie_id` RIGHT JOIN `productors` AS `tab4` ON `tab3`.`id` = `tab4`.`user_id`');
    $this->assertEquals([['tab1_date' => null,
                          'tab2_name' => null,
                          'tab3_libelle' => null,
                          'tab4_label' => 'Productor 1',
                          'tab4_birthday' => '01/01/2015'],
                         ['tab1_date' => null,
                          'tab2_name' => null,
                          'tab3_libelle' => null,
                          'tab4_label' => 'Productor 2',
                          'tab4_birthday' => '02/02/2016'],
                         ['tab1_date' => '2018',
                          'tab2_name' => 'Actor name 3',
                          'tab3_libelle' => 'User 3',
                          'tab4_label' => 'Productor 3',
                          'tab4_birthday' => '03/03/2017'],
                         ['tab1_date' => '2018',
                          'tab2_name' => 'Actor name 3',
                          'tab3_libelle' => 'User 3',
                          'tab4_label' => 'Productor 4',
                          'tab4_birthday' => '04/04/2018']],
                        array_map(fn($row) => $row->toArray(),
                                  $results));
    $this->assertTrue($results[0] instanceof Storm_Model_Row);
  }
}




class Storm_Test_LoaderQueryVolatileInSubQueriesTest
  extends Storm_Test_ModelTestCase {

  protected array $_storm_scopes = ['LoaderQueryVolatileInSubQueriesTest'];

  public function setUp(): void
  {
    parent::setUp();

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 1,
                    'name' => 'User name 1',
                    'year' => '2020',
                    'user_id' => 1]);
    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 2,
                    'name' => 'User name 2',
                    'year' => '2022',
                    'user_id' => 2]);
    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 3,
                    'name' => 'User name 3',
                    'year' => '2021',
                    'user_id' => 3]);
    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 4,
                    'name' => 'User name 4',
                    'year' => '2018',
                    'user_id' => 4]);
  }


  /** @test */
  public function withInSubQueriesReturnUser_Name3_Name4() {
    $results = Storm_Test_LoaderQueryUser::query()
      ->in('user_id', Storm_Test_LoaderQueryUser::subIn('users2')
           ->select(['user_id'])
           ->in('year', ['2021', '2018']))
      ->fetchAll();

    $this->assertCount(2, $results);
    $this->assertEquals(['User name 3', 'User name 4'],
                        array_map(fn($row) => $row->getName(),
                                  $results));
    $this->assertTrue($results[0] instanceof Storm_Test_LoaderQueryUser);
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`user_id` IN (SELECT `users2`.`user_id` FROM `users` AS `users2` WHERE (`users2`.`year` IN (\'2021\', \'2018\'))))');
  }
}




class Storm_Test_LoaderQueryVolatileNotInSubQueriesTest
  extends Storm_Test_ModelTestCase {

  protected array $_storm_scopes = ['LoaderQueryVolatileNotInSubQueriesTest'];

  public function setUp(): void
  {
    parent::setUp();

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 1,
                    'name' => 'User name 1',
                    'actor_id' => 1]);
    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 2,
                    'name' => 'User name 2',
                    'actor_id' => 2]);
    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 3,
                    'name' => 'User name 3',
                    'actor_id' => 3]);

    $this->fixture(Storm_Test_LoaderQueryActor::class,
                   ['id' => 2,
                    'id_actor' => 2,
                    'label' => 'Actor 2']);
    $this->fixture(Storm_Test_LoaderQueryActor::class,
                   ['id' => 3,
                    'id_actor' => 3,
                    'label' => 'Actor 3']);
  }


  /** @test */
  public function withNotInSubQueriesReturnUser_Name1_Name2() {
    $results = Storm_Test_LoaderQueryUser::query()
      ->not_in('actor_id', Storm_Test_LoaderQueryActor::subIn()
               ->select(['id_actor'])
               ->eq('label', 'Actor 3'))
      ->fetchAll();

    $this->assertCount(2, $results);
    $this->assertEquals(['id' => 1,
                         'name' => 'User name 1',
                         'actor_id' => 1], $results[0]->toArray());
    $this->assertEquals(['id' => 2,
                         'name' => 'User name 2',
                         'actor_id' => 2], $results[1]->toArray());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (`users`.`actor_id` NOT IN (SELECT `actors`.`id_actor` FROM `actors` WHERE (`actors`.`label` = \'Actor 3\')))');
  }
}




class Storm_Test_LoaderQueryVolatileExistsSubQueriesTest
  extends Storm_Test_ModelTestCase {

  protected array $_storm_scopes = ['LoaderQueryVolatileExistsSubQueriesTest'];

  public function setUp(): void
  {
    parent::setUp();

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 1,
                    'name' => 'User name 1',
                    'actor_id' => 1,
                    'age' => 21]);
    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 2,
                    'name' => 'User name 2',
                    'actor_id' => 2,
                    'age' => 22]);
    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 3,
                    'name' => 'User name 3',
                    'actor_id' => 3,
                    'age' => 23]);

    $this->fixture(Storm_Test_LoaderQueryActor::class,
                   ['id' => 2,
                    'id_actor' => 2,
                    'label' => 'Actor 2',
                    'age' => 22]);
    $this->fixture(Storm_Test_LoaderQueryActor::class,
                   ['id' => 3,
                    'id_actor' => 3,
                    'label' => 'Actor 3',
                    'age' => 20]);
  }


  /** @test */
  public function withExistsSubQueriesReturnOnlyUser_Name3() {
    $results = Storm_Test_LoaderQueryUser::query()
      ->exists(Storm_Test_LoaderQueryActor::subExists()
               ->on_eq('actor_id', 'id_actor')
               ->on_gt('age', 'age'))
      ->fetchAll();

    $this->assertCount(1, $results);
    $this->assertEquals(['id' => 3,
                         'name' => 'User name 3',
                         'actor_id' => 3,
                         'age' => 23], $results[0]->toArray());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (EXISTS (SELECT \'x\' FROM `actors` WHERE (`users`.`actor_id` = `actors`.`id_actor` AND `users`.`age` > `actors`.`age`)))');
  }


  /** @test */
  public function withExistsSubQueriesReturnOnlyUser_Name2() {
    $results = Storm_Test_LoaderQueryUser::query()
      ->exists(Storm_Test_LoaderQueryActor::subExists()
               ->on_gt_eq('actor_id', 'id_actor')
               ->on_lt_eq('age', 'age'))
      ->fetchAll();

    $this->assertCount(1, $results);
    $this->assertEquals(['id' => 2,
                         'name' => 'User name 2',
                         'actor_id' => 2,
                         'age' => 22], $results[0]->toArray());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (EXISTS (SELECT \'x\' FROM `actors` WHERE (`users`.`actor_id` >= `actors`.`id_actor` AND `users`.`age` <= `actors`.`age`)))');
  }


  /** @test */
  public function withNotExistsSubQueriesWithTwoConditionReturnUser_Name2_Name3() {
    $results = Storm_Test_LoaderQueryUser::query()
      ->not_exists(Storm_Test_LoaderQueryActor::subExists()
                   ->on_not_eq('actor_id', 'id_actor')
                   ->on_lt('age', 'age'))
      ->fetchAll();

    $this->assertCount(2, $results);
    $this->assertEquals(['id' => 2,
                         'name' => 'User name 2',
                         'actor_id' => 2,
                         'age' => 22], $results[0]->toArray());
    $this->assertEquals(['id' => 3,
                         'name' => 'User name 3',
                         'actor_id' => 3,
                         'age' => 23], $results[1]->toArray());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (NOT EXISTS (SELECT \'x\' FROM `actors` WHERE (`users`.`actor_id` != `actors`.`id_actor` AND `users`.`age` < `actors`.`age`)))');
  }


  /** @test */
  public function withNotExistsSubQueriesWithOneConditionAndOneCriteriaReturnUser_Name1_Name2() {
    $results = Storm_Test_LoaderQueryUser::query()
      ->not_exists(Storm_Test_LoaderQueryActor::subExists()
                   ->on_eq('actor_id', 'id_actor')
                   ->lt('age', 22))
      ->fetchAll();

    $this->assertCount(2, $results);
    $this->assertEquals(['id' => 1,
                         'name' => 'User name 1',
                         'actor_id' => 1,
                         'age' => 21], $results[0]->toArray());
    $this->assertEquals(['id' => 2,
                         'name' => 'User name 2',
                         'actor_id' => 2,
                         'age' => 22], $results[1]->toArray());
    $this->assertSql('SELECT `users`.* FROM `users` WHERE (NOT EXISTS (SELECT \'x\' FROM `actors` WHERE (`users`.`actor_id` = `actors`.`id_actor` AND `actors`.`age` < 22)))');
  }


  /** @test */
  public function withExistsInCriteriaOrShouldHaveSql() {
    Storm_Test_LoaderQueryUser::query()
      ->or((new Storm_Query_Criteria)
           ->eq('age', 22)
           ->exists(Storm_Test_LoaderQueryActor::subExists()
                    ->on_eq('actor_id', 'id_actor')
                    ->eq('age', 20)))
      ->fetchAll();

    $this->assertSql('SELECT `users`.* FROM `users` WHERE ((`users`.`age` = 22 OR EXISTS (SELECT \'x\' FROM `actors` WHERE (`users`.`actor_id` = `actors`.`id_actor` AND `actors`.`age` = 20))))');
  }
}




class Storm_Test_LoaderQueryWithErrorsTest
  extends Storm_Test_ModelTestCase {

  protected array $_storm_scopes = ['Storm_Test_LoaderQueryWithErrorsTest'];

  public function setUp(): void
  {
    parent::setUp();

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 1,
                    'name' => 'User name 1',
                    'actor_id' => 1,
                    'age' => 21]);

    $this->fixture(Storm_Test_LoaderQueryActor::class,
                   ['id' => 2,
                    'id_actor' => 1,
                    'label' => 'Actor 2',
                    'age' => 22]);
  }


  /** @test */
  public function existsWithNoConditionAndSelectColumnReturnNothingAndSqlContainErrorMessage() {
    try {
      Storm_Test_LoaderQueryUser::query()
        ->exists(Storm_Test_LoaderQueryActor::subExists()
                 ->select(['error']))
        ->fetchAll();
    } catch (Exception $e) {
      $this->assertSql('Error: For Exists don\'t add select column, Error: For Exists condition "on" is mandatory');
      return;
    }

    $this->assertTrue(false);
  }


  /** @test */
  public function inWithNoSelectColumnReturnNothingAndSqlContainErrorMessage() {
    try {
      $results = Storm_Test_LoaderQueryUser::query()
        ->in('actor_id', Storm_Test_LoaderQueryActor::subIn())
        ->fetchAll();
    } catch (Exception $e) {
      $this->assertSql('Error: For In add select column is mandatory');
      return;
    }

    $this->assertTrue(false);
  }


  /** @test */
  public function inWithMultipleSelectColumnReturnNothingAndSqlContainErrorMessage() {
    try {
      $results = Storm_Test_LoaderQueryUser::query()
        ->in('actor_id', Storm_Test_LoaderQueryActor::subIn()
             ->select(['id', 'id_actor']))
        ->fetchAll();
    } catch (Exception $e) {
      $this->assertSql('Error: For In add select must have only one column');
      return;
    }

    $this->assertTrue(false);
  }


  /** @test */
  public function innerWithNoConditionReturnNothingAndSqlContainErrorMessage() {
    try {
      $results = Storm_Test_LoaderQueryUser::join('tab1')
        ->all_select()
        ->inner(Storm_Test_LoaderQueryActor::inner('tab2'))
        ->fetchAll();
    } catch (Exception $e) {
      $this->assertSql('Error: For Inner condition "on" is mandatory');
      return;
    }

    $this->assertTrue(false);
  }


  /** @test */
  public function innerWithNoSelectReturnNothingAndSqlContainErrorMessage() {
    try {
      $results = Storm_Test_LoaderQueryUser::join('tab1')
        ->inner(Storm_Test_LoaderQueryActor::inner('tab2')
                ->on_eq('actor_id', 'id_actor'))
        ->fetchAll();
    } catch (Exception $e) {
      $this->assertSql('Error: For Join a selection is mandatory');
      return;
    }

    $this->assertTrue(false);
  }


  /** @test */
  public function leftWithNoConditionReturnNothingAndSqlContainErrorMessage() {
    try {
      $results = Storm_Test_LoaderQueryUser::join('tab1')
        ->all_select()
        ->left(Storm_Test_LoaderQueryActor::left('tab2'))
        ->fetchAll();
    } catch (Exception $e) {
      $this->assertSql('Error: For Left condition "on" is mandatory');
      return;
    }

    $this->assertTrue(false);
  }


  /** @test */
  public function rightWithNoConditionReturnNothingAndSqlContainErrorMessage() {
    try {
      $results = Storm_Test_LoaderQueryUser::join('tab1')
        ->all_select()
        ->right(Storm_Test_LoaderQueryActor::right('tab2'))
        ->fetchAll();
    } catch (Exception $e) {
      $this->assertSql('Error: For Right condition "on" is mandatory');
      return;
    }

    $this->assertTrue(false);
  }
}




class Storm_Test_LoaderQueryUpdateTest extends Storm_Test_ModelTestCase {

  protected array $_storm_scopes = ['Storm_Test_LoaderQueryUpdateTest'];

  public function setUp(): void
  {
    parent::setUp();

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 1,
                    'name' => 'User name 1',
                    'age' => 21,
                    'clean_field' => ' field    to be  cleaned   ']);

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 2,
                    'name' => 'Users names 2',
                    'age' => 22,
                    'clean_field' => 'field dont need clean']);
  }


  /** @test */
  public function withOnlyOneSetShouldReturnUpdateQueryAndUpdateVolatile() {
    $count = Storm_Test_LoaderQueryUser::updateQuery()
      ->addSet('name', 'username 2 updated')
      ->match((new Storm_Query_MatchBoolean('age'))
              ->against_or(22))
      ->updateAll();

    Storm_Test_LoaderQueryUser::clearCache();

    $this->assertEquals(1, $count);
    $this->assertEquals([['id' => 1,
                          'name' => 'User name 1',
                          'age' => 21,
                          'clean_field' => ' field    to be  cleaned   '],
                         ['id' => 2,
                          'name' => 'username 2 updated',
                          'age' => 22,
                          'clean_field' => 'field dont need clean']],
                        array_map(fn($user) => $user->attributesToArray(),
                                  Storm_Test_LoaderQueryUser::findAllBy([])));
    $this->assertSql('UPDATE `users` SET `name` = \'username 2 updated\' WHERE (MATCH(`users`.`age`) AGAINST(\'22\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function queryPromoteUpdateWithOnlyOneSetShouldReturnUpdateQueryAndUpdateVolatile() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->select('age')
      ->order('age')
      ->match((new Storm_Query_MatchBoolean('age'))
              ->against_and(22));
    $update = $query->promoteToUpdate();

    $count = $update->addSet('name', 'username 2 updated')
                    ->updateAll();

    Storm_Test_LoaderQueryUser::clearCache();

    $this->assertEquals(1, $count);
    $this->assertEquals([['id' => 1,
                          'name' => 'User name 1',
                          'age' => 21,
                          'clean_field' => ' field    to be  cleaned   '],
                         ['id' => 2,
                          'name' => 'username 2 updated',
                          'age' => 22,
                          'clean_field' => 'field dont need clean']],
                        array_map(fn($user) => $user->attributesToArray(),
                                  Storm_Test_LoaderQueryUser::findAllBy([])));
    $this->assertSql('UPDATE `users` SET `name` = \'username 2 updated\' WHERE (MATCH(`users`.`age`) AGAINST(\'+22\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function withTwoSetsShouldReturnUpdateQueryAndUpdateVolatile() {
    $count = Storm_Test_LoaderQueryUser::updateQuery()
      ->addSet('name', 'username 2 updated')
      ->addSet('age', 35)
      ->eq('age', 22)
      ->updateAll();

    Storm_Test_LoaderQueryUser::clearCache();

    $this->assertEquals(1, $count);
    $this->assertEquals([['id' => 1,
                          'name' => 'User name 1',
                          'age' => 21,
                          'clean_field' => ' field    to be  cleaned   '],
                         ['id' => 2,
                          'name' => 'username 2 updated',
                          'age' => 35,
                          'clean_field' => 'field dont need clean']],
                        array_map(fn($user) => $user->attributesToArray(),
                                  Storm_Test_LoaderQueryUser::findAllBy([])));
    $this->assertSql('UPDATE `users` SET `name` = \'username 2 updated\', `age` = 35 WHERE (`users`.`age` = 22)');
  }


  /** @test */
  public function withCleanSpacesShouldReturnUpdateQueryAndUpdateVolatile() {
    $count = Storm_Test_LoaderQueryUser::updateQuery()
      ->addSet(Storm_Query_Key::cleanSpaces('clean_field'))
      ->updateAll();

    Storm_Test_LoaderQueryUser::clearCache();

    $this->assertEquals(2, $count);
    $this->assertEquals([['id' => 1,
                          'name' => 'User name 1',
                          'age' => 21,
                          'clean_field' => 'field to be cleaned'],
                         ['id' => 2,
                          'name' => 'Users names 2',
                          'age' => 22,
                          'clean_field' => 'field dont need clean']],
                        array_map(fn($user) => $user->attributesToArray(),
                                  Storm_Test_LoaderQueryUser::findAllBy([])));
    $this->assertSql("UPDATE `users` SET `clean_field` = CLEAN_SPACES(`clean_field`)");
  }


  /** @test */
  public function withConcatShouldReturnUpdateQueryAndUpdateVolatile() {
    $count = Storm_Test_LoaderQueryUser::updateQuery()
      ->addSet(Storm_Query_Key::concat('name', ' with concat'))
      ->updateAll();

    Storm_Test_LoaderQueryUser::clearCache();

    $this->assertEquals(2, $count);
    $this->assertEquals([['id' => 1,
                          'name' => 'User name 1 with concat',
                          'age' => 21,
                          'clean_field' => ' field    to be  cleaned   '],
                         ['id' => 2,
                          'name' => 'Users names 2 with concat',
                          'age' => 22,
                          'clean_field' => 'field dont need clean']],
                        array_map(fn($user) => $user->attributesToArray(),
                                  Storm_Test_LoaderQueryUser::findAllBy([])));
    $this->assertSql("UPDATE `users` SET `name` = CONCAT(`name`, ' with concat')");
  }


  /** @test */
  public function withRegexReplaceShouldReturnUpdateQueryAndUpdateVolatile() {
    Storm_Test_LoaderQueryUser::find(1)->setName('http://url.com')->assertSave();
    Storm_Test_LoaderQueryUser::find(2)->setName('url.com')->assertSave();
    $count = Storm_Test_LoaderQueryUser::updateQuery()
      ->addSet(Storm_Query_Key::regexpReplace('name', 'http[s]*://.*', 'https://new_url.com'))
      ->updateAll();

    Storm_Test_LoaderQueryUser::clearCache();

    $this->assertEquals(2, $count);
    $this->assertEquals([['id' => 1,
                          'name' => 'https://new_url.com',
                          'age' => 21,
                          'clean_field' => ' field    to be  cleaned   '],
                         ['id' => 2,
                          'name' => 'url.com',
                          'age' => 22,
                          'clean_field' => 'field dont need clean']],
                        array_map(fn($user) => $user->attributesToArray(),
                                  Storm_Test_LoaderQueryUser::findAllBy([])));
    $this->assertSql("UPDATE `users` SET `name` = REGEXP_REPLACE(`name`, 'http[s]*://.*', 'https://new_url.com')");
  }


  /** @test */
  public function withRegexReplaceCleanSpacesShouldReturnUpdateQueryAndUpdateVolatile() {
    $count = Storm_Test_LoaderQueryUser::updateQuery()
      ->addSet(Storm_Query_Key::regexpReplace('clean_field', '\\bto\\s+be\\b', 'has been')
               ->addCleanSpaces())
      ->updateAll();

    Storm_Test_LoaderQueryUser::clearCache();

    $this->assertEquals(2, $count);
    $this->assertEquals([['id' => 1,
                          'name' => 'User name 1',
                          'age' => 21,
                          'clean_field' => 'field has been cleaned'],
                         ['id' => 2,
                          'name' => 'Users names 2',
                          'age' => 22,
                          'clean_field' => 'field dont need clean']],
                        array_map(fn($user) => $user->attributesToArray(),
                                  Storm_Test_LoaderQueryUser::findAllBy([])));
    $this->assertSql("UPDATE `users` SET `clean_field` = CLEAN_SPACES(REGEXP_REPLACE(`clean_field`, '\\\\bto\\\\s+be\\\\b', 'has been'))");
  }
}




class Storm_Test_LoaderQueryDeleteTest extends Storm_Test_ModelTestCase {

  protected array $_storm_scopes = ['Storm_Test_LoaderQueryDeleteTest'];

  public function setUp(): void
  {
    parent::setUp();

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 1,
                    'name' => 'User name 1',
                    'age' => 21]);

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 2,
                    'name' => 'User name 2',
                    'age' => 22]);
  }


  /** @test */
  public function onDeleteBasicShouldReturnDeleteQueryAndDeleteVolatile() {
    $count = Storm_Test_LoaderQueryUser::deleteQuery()
      ->beBasic()
      ->match((new Storm_Query_MatchBoolean('age'))
              ->against_or(22))
      ->deleteAll();

    Storm_Test_LoaderQueryUser::clearCache();

    $this->assertEquals(1, $count);
    $this->assertEquals([['id' => 1,
                          'name' => 'User name 1',
                          'age' => 21]],
                        array_map(fn($user) => $user->attributesToArray(),
                                  Storm_Test_LoaderQueryUser::findAllBy([])));
    $this->assertSql('DELETE FROM `users` WHERE (MATCH(`users`.`age`) AGAINST(\'22\' IN BOOLEAN MODE))');
  }


  /** @test */
  public function onDeleteNonBasicShouldReturnSelectQueryAndDeleteVolatile() {
    $count = Storm_Test_LoaderQueryUser::deleteQuery()
      ->match((new Storm_Query_MatchBoolean('age'))
              ->against_or(22))
      ->deleteAll();

    Storm_Test_LoaderQueryUser::clearCache();

    $this->assertEquals(1, $count);
    $this->assertEquals([['id' => 1,
                          'name' => 'User name 1',
                          'age' => 21]],
                        array_map(fn($user) => $user->attributesToArray(),
                                  Storm_Test_LoaderQueryUser::findAllBy([])));
    $this->assertSqlCount('SELECT `users`.* FROM `users` WHERE (MATCH(`users`.`age`) AGAINST(\'22\' IN BOOLEAN MODE)) LIMIT 100', 1);
  }


  /** @test */
  public function onDeleteAllUsersMoreThan100NonBasicShouldReturnSelectQueryAndDeleteVolatile() {
    foreach (range(100, 200) as $id)
      $this->fixture(Storm_Test_LoaderQueryUser::class,
                     ['id' => $id,
                      'name' => 'User name ' . $id,
                      'age' => $id]);

    $this->assertEquals(103, Storm_Test_LoaderQueryUser::countBy([]));
    $count = Storm_Test_LoaderQueryUser::deleteQuery()
      ->deleteAll();

    Storm_Test_LoaderQueryUser::clearCache();

    $this->assertSqlCount("SELECT `users`.* FROM `users` LIMIT 100", 2);
    $this->assertEquals(103, $count);
  }
}




class Storm_Test_LoaderQueryBasicSelectTest extends Storm_Test_ModelTestCase {

  protected array $_storm_scopes = ['Storm_Test_LoaderQueryBasicSelectTest'];

  public function setUp(): void
  {
    parent::setUp();

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 1,
                    'name' => 'User name 1',
                    'age' => 21]);

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 2,
                    'name' => 'User name 2',
                    'age' => 22]);
  }


  /** @test */
  public function withModeBasicShouldReturnBasicArray() {
    $array_result = Storm_Test_LoaderQueryUser::query()
      ->select(['id', 'name'])
      ->beBasic()
      ->fetchAll();

    $this->assertEquals([['id' => 1,
                          'name' => 'User name 1'],
                         ['id' => 2,
                          'name' => 'User name 2']],
                        $array_result);
    $this->assertSql('SELECT `users`.`id`, `users`.`name` FROM `users`');
  }
}




class Storm_Test_LoaderQueryCloneTest extends Storm_Test_ModelTestCase {

  protected array $_storm_scopes = ['Storm_Test_LoaderQueryCloneTest'];

  public function setUp(): void
  {
    parent::setUp();

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 1,
                    'name' => 'User name',
                    'age' => 21]);

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 2,
                    'name' => 'User name',
                    'age' => 22]);
  }


  /** @test */
  public function withCloneShouldReturnAllUsers() {
    $query = Storm_Test_LoaderQueryUser::query()
      ->gt('age', 21);

    $clone_select = $query->clone()->select('id');

    $clone_order = $query->clone('orders');

    $query->fetchAll();
    $clone_select->fetchFirst();
    $clone_order->countAll();

    $this->assertSqlEquals(['SELECT `users`.* FROM `users` WHERE (`users`.`age` > 21)',
                            'SELECT `users`.`id` FROM `users` WHERE (`users`.`age` > 21) LIMIT 1',
                            'SELECT COUNT(*) AS `numberof` FROM `users` AS `orders` WHERE (`orders`.`age` > 21)']);
  }
}




class Storm_Test_LoaderQueryCountTest extends Storm_Test_ModelTestCase {

  protected array $_storm_scopes = ['Storm_Test_LoaderQueryCountTest'];

  public function setUp(): void
  {
    parent::setUp();

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 1,
                    'name' => 'User name 1']);

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 2,
                    'name' => 'User name 2']);
  }


  /** @test */
  public function withQueryCountShouldReturnCount_Star() {
    $count = Storm_Test_LoaderQueryUser::query()
      ->countAll();

    $this->assertEquals(2, $count);
    $this->assertSql('SELECT COUNT(*) AS `numberof` FROM `users`');
  }
}




class Storm_Test_LoaderQueryFunctionDateTest extends Storm_Test_ModelTestCase
{

  protected array $_storm_scopes = ['Storm_Test_LoaderQueryFunctionDateTest'];

  public function setUp(): void
  {
    parent::setUp();

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 1,
                    'name' => 'User name 1',
                    'date_maj' => '2024-10-01 10:00:00']);

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 2,
                    'name' => 'User name 2',
                    'date_maj' => '2024-10-01']);

    $this->fixture(Storm_Test_LoaderQueryUser::class,
                   ['id' => 3,
                    'name' => 'User name 3',
                    'date_maj' => '2024-10-02 10:00:00']);
  }


  /** @test */
  public function selectWithDate2024_10_01ShouldReturnUserId_1_2()
  {
    $results = Storm_Test_LoaderQueryUser::query()
      ->eq(Storm_Query_Key::date('date_maj'), '2024-10-01')
      ->fetchAll();

    $this->assertEquals([['id' => 1,
                          'name' => 'User name 1'],
                         ['id' => 2,
                          'name' => 'User name 2']],
                        array_map(fn($res) => ['id' => $res->getId(), 'name' => $res->getName()],
                                  $results));
    $this->assertSql("SELECT `users`.* FROM `users` WHERE (DATE(`users`.`date_maj`) = '2024-10-01')");
  }
}




class Storm_Test_LoaderQueryUser extends Storm_Model_Abstract {

  protected $_table_name = 'users';
}




class Storm_Test_LoaderQueryMovie extends Storm_Model_Abstract {

  protected $_table_name = 'movies';
}




class Storm_Test_LoaderQueryActor extends Storm_Model_Abstract {

  protected $_table_name = 'actors';
}




class Storm_Test_LoaderQueryProductor extends Storm_Model_Abstract {

  protected $_table_name = 'productors';
}
